/*
    Copyright (c) 2013 Boldizsár Lipka <lipkab@zoho.com>

    This file is part of the Apertium Visual Rule Editor.

    Apertium Visual Rule Editor is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Apertium Visual Rule Editor is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Apertium Visual Rule Editor.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef CONFIG_H
#define CONFIG_H


#include <QColor>
#include <QXmlDefaultHandler>
#include <QFont>
#include "node.h"

namespace reptype
{
enum Type
{
    BOX = 0,
    TAB,
    HIDDEN,
    UNKNOWN
};
}

namespace contype
{
enum Type
{
    NESTED = 0,
    ARROW,
    UNKNOWN
};
}

namespace proptype
{
enum Type
{
    STRING = 0,
    INTEGER,
    SELECTION,
    UNKNOWN
};
}

class VisualSchema
{
public:
    VisualSchema(const QXmlInputSource& file);
    VisualSchema();

    struct Tag
    {
        reptype::Type reptype;
        QColor boxColor;
        QString label;
        QStringList children;
        QStringList properties;
        QString nesting;
        QString name;
    };

    struct Property
    {
        proptype::Type type;
        bool isMandatory;
        QString label;
        QString valueListId;
        QString fullName;
    };

    struct Connection
    {
        contype::Type type;
    };

    const Tag tag(const QString& name) const { return tags_[name]; }
    const Property property(const QString& tagname, const QString& propname) const { return props_[tagname + "/" + propname]; }
    const Property property(const QString& tagperprop) const { return props_[tagperprop]; }
    const Connection connection(const QString& parent, const QString& child) const { return cons_[parent + "/" + child]; }

private:
    class XmlHandler : public QXmlDefaultHandler
    {
    public:
        XmlHandler(VisualSchema* vs)
            : QXmlDefaultHandler()
            , vschema_(vs) {}

        virtual bool startElement(const QString &namespaceURI, const QString &localName, const QString &qName, const QXmlAttributes &atts);
        virtual bool endElement(const QString &namespaceURI, const QString &localName, const QString &qName);

    private:
        static reptype::Type toRepType(const QString& str);
        static contype::Type toConType(const QString& str);
        static proptype::Type toPropType(const QString& str);
        static QColor toColor(const QString& str);
        static bool toBool(const QString& str) { if (str == "true") return true; return false; }

        Tag currentTag_;
        QString currentTagName_;
        VisualSchema* vschema_;
    };

    QMap<QString, Tag> tags_;
    QMap<QString, Property> props_;
    QMap<QString, Connection> cons_;
};

class ValueMap
{
public:
    ValueMap(const QXmlInputSource& file);
    ValueMap();

    QMap<QString, QString> list(const QString& id) const { return lists_[id]; }
    QMap<QString, QString> rlist(const QString& id) const { return rlists_[id]; }

private:
    class XmlHandler : public QXmlDefaultHandler
    {
    public:
        XmlHandler(ValueMap* vl)
            : vmap_(vl)
            , currentListId_()
        {}

        virtual bool startElement(const QString &namespaceURI, const QString &localName, const QString &qName, const QXmlAttributes &atts);

    private:
        ValueMap* vmap_;
        QString currentListId_;
    };

    QMap<QString, QMap<QString, QString> > lists_;
    QMap<QString, QMap<QString, QString> > rlists_;
};

class Configuration;

const Configuration& appConfig();

class Configuration
{
    friend const Configuration& appConfig();
public:
    const VisualSchema& visualSchema() const { return vschema_; }
    const ValueMap& valueMap() const { return vmap_; }
    const QMap<QString,QString>& templates() const { return templates_; }

private:
    Configuration();

    class TemplatesXmlHandler : public QXmlDefaultHandler
    {
    public:
        TemplatesXmlHandler(QMap<QString,QString>* map)
            : map_(map)
        {}

        virtual bool startElement(const QString &namespaceURI, const QString &localName, const QString &qName, const QXmlAttributes &atts);

    private:
        QMap<QString,QString>* map_;
    };

    VisualSchema vschema_;
    ValueMap vmap_;
    QMap<QString, QString> templates_;
};

namespace appearance
{
inline QColor fontColor() { return Qt::black; }
QString formatTextNormal(const QString& str);
QString formatTextBold(const QString& str);
}

#endif // CONFIG_H
