/*
    Copyright (c) 2013 Boldizsár Lipka <lipkab@zoho.com>

    This file is part of the Apertium Visual Rule Editor.

    Apertium Visual Rule Editor is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Apertium Visual Rule Editor is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Apertium Visual Rule Editor.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <QFileDialog>
#include <QDebug>
#include <QTableWidget>
#include <QLabel>
#include <QXmlInputSource>
#include <QCloseEvent>
#include <QMessageBox>

#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "filetab.h"
#include "newfiledialog.h"
#include "resources.h"

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    newFileDialog_(this),
    stack_(NULL),
    files_(NULL),
    ui(new Ui::MainWindow)
{
    connect(&newFileDialog_, SIGNAL(accepted()), this, SLOT(createNewFile()));
    ui->setupUi(this);

    files_ = findChild<QTabWidget*>("filesContainer");
    connect(files_, SIGNAL(currentChanged(int)), this, SLOT(updateActionStack()));
    connect(files_, SIGNAL(currentChanged(int)), this, SLOT(updateActionStack()));
    updateActionStack();
}

MainWindow::~MainWindow()
{
    delete ui;
}

void MainWindow::on_actionOpen_activated()
{
    QString fileName = QFileDialog::getOpenFileName(this, tr("Open file"), "", tr("Apertium transfer rules (*.t1x *.t2x *.t3x)"));

    if (fileName.isEmpty()) {
        return;
    }

    QFile file(fileName);
    QXmlInputSource in(&file);
    FileTab* ft = new FileTab(this, readXmlIntoNode(in), fileName);

    connect(ft, SIGNAL(sectionChanged()), this, SLOT(updateActionStack()));
    connect(ft, SIGNAL(sectionChanged()), this, SLOT(updateSidebar()));
    connect(ft, SIGNAL(saveStateChanged(bool)), this, SLOT(updateFileTabLabels()));
    files_->addTab(ft, ft->fileName());
    files_->setCurrentIndex(files_->count() - 1);
}

void MainWindow::on_actionClose_activated()
{
    closeFile(files_->currentIndex());
}

void MainWindow::on_actionSave_activated()
{
    saveFile(files_->currentIndex());
}

void MainWindow::on_actionSaveAs_activated()
{
    saveFileAs(files_->currentIndex());
}

void MainWindow::on_actionNew_activated()
{
    newFileDialog_.open();
}

void MainWindow::on_actionUndo_activated()
{
    stack_->undo();
}

void MainWindow::on_actionRedo_activated()
{
    stack_->redo();
}

void MainWindow::updateUndoRedo()
{
    QAction* undo = findChild<QAction*>("actionUndo");
    QAction* redo = findChild<QAction*>("actionRedo");

    if (stack_ == NULL) {
        undo->setEnabled(false);
        redo->setEnabled(false);
    } else {
        undo->setEnabled(stack_->canUndo());
        redo->setEnabled(stack_->canRedo());
    }
}

void MainWindow::updateActionStack()
{
    if (stack_ != NULL) {
        disconnect(stack_, SIGNAL(actionPerformed()), this, SLOT(updateUndoRedo()));
    }

    stack_ = currentActionStack();

    if (stack_ != NULL) {
        connect(stack_, SIGNAL(actionPerformed()), this, SLOT(updateUndoRedo()));
    }

    updateUndoRedo();
}

ActionStack *MainWindow::currentActionStack()
{
    QWidget* w = files_->currentWidget();

    if (w == NULL) {
        return NULL;
    }
    FileTab* ft = static_cast<FileTab*>(w);
    return ft->currentActionStack();
}

void MainWindow::createNewFile()
{
    Node *root = NULL;

    QString path = newFileDialog_.getChoice();

    QFile file(path);
    QXmlInputSource input(&file);
    root = readXmlIntoNode(input);

    FileTab* ft = new FileTab(this, root, "");
    ft->setUnsaved();
    connect(ft, SIGNAL(sectionChanged()), this, SLOT(updateActionStack()));
    connect(ft, SIGNAL(saveStateChanged(bool)), this, SLOT(updateFileTabLabels()));
    files_->addTab(ft, ft->fileName() + "*");
    files_->setCurrentIndex(files_->count() - 1);
    updateSidebar();
}

void MainWindow::updateFileTabLabels()
{
    for (int i = 0; i<files_->count(); i++) {
        FileTab* ft = static_cast<FileTab*>(files_->widget(i));
        QString label = ft->fileName();
        label += ft->isSaved() ? "" : "*";
        files_->setTabText(i, label);
    }
}

void MainWindow::updateSidebar()
{
    FileTab* ft = tab(files_->currentIndex());
    SectionTab* st = ft->currentSection();

    if (st->selectedBox() == NULL) {
        boxbar()->clearItems();
        propertybar()->clearItems();
        return;
    }

    VisualSchema::Tag tdef = appConfig().visualSchema().tag(st->selectedBox()->data()->name());
    boxbar()->clearItems();
    foreach (QString str, tdef.children) {
        VisualSchema::Tag cdef = appConfig().visualSchema().tag(str);
        boxbar()->addItem(new BoxPreview(cdef));
    }

    propertybar()->clearItems();
    foreach (QString str, tdef.properties) {
        VisualSchema::Property pdef = appConfig().visualSchema().property(str);
        propertybar()->addItem(new PropertyPreview(pdef));
    }
}

void MainWindow::closeEvent(QCloseEvent *ev)
{
    if (!closeAllFiles()) {
        ev->ignore();
    }
}

void MainWindow::saveFile(int index)
{
    FileTab* ft = tab(index);

    if (!ft->filePath().isEmpty()) {
        QFile dst(ft->filePath());
        dst.open(QIODevice::WriteOnly | QIODevice::Text);
        QTextStream out(&dst);

        out << ft->rootNode()->toXml();
        out.flush();
        dst.close();
        ft->setSaved();
    } else {
        saveFileAs(index);
    }
}

void MainWindow::saveFileAs(int index)
{
    FileTab* ft =tab(index);

    QString path = QFileDialog::getSaveFileName(this, tr("Save as"), "", tr("Apertium transfer rules (*.t1x *.t2x *.t3x)"));
    if (path.isEmpty()) {
        return;
    }

    QFile dst(path);
    dst.open(QIODevice::WriteOnly | QIODevice::Text);
    QTextStream out(&dst);

    out << ft->rootNode()->toXml();
    ft->setSaved();

    ft->setFilePath(path);
    files_->setTabText(files_->currentIndex(), ft->fileName());
}

bool MainWindow::closeFile(int index)
{
    FileTab* ft = tab(index);

    if (ft->isSaved()) {
        files_->removeTab(index);
        return true;
    }

    QMessageBox msg(this);
    msg.setIcon(QMessageBox::Warning);
    msg.setText(tr("This file has unsaved changes."));
    msg.setInformativeText(tr("Do you want to save before closing?"));
    msg.setStandardButtons(QMessageBox::Cancel | QMessageBox::Save | QMessageBox::Discard);

    int res = msg.exec();

    switch (res) {
    case QMessageBox::Save:
        saveFile(index);
    case QMessageBox::Discard:
        files_->removeTab(index);
        return true;
    case QMessageBox::Cancel:
    default:
        return false;
    }
}

bool MainWindow::closeAllFiles()
{
    bool needConfirm = false;
    for (int i = 0; i<files_->count(); i++) {
        if (!tab(i)->isSaved()) {
            needConfirm = true;
            break;
        }
    }

    if (!needConfirm) {
        while (files_->count() > 0) { files_->removeTab(0); }
        return true;
    }

    QMessageBox msg(this);
    msg.setIcon(QMessageBox::Warning);
    msg.setText(tr("Some files have unsaved changes."));
    msg.setInformativeText(tr("Do you want to save before closing?"));
    msg.setStandardButtons(QMessageBox::Cancel | QMessageBox::SaveAll | QMessageBox::Discard);

    int res = msg.exec();

    switch (res) {
    case QMessageBox::SaveAll:
        for (int i = 0; i<files_->count(); i++) { saveFile(i); }
    case QMessageBox::Discard:
        while (files_->count() > 0) { files_->removeTab(0); }
        return true;
    case QMessageBox::Cancel:
    default:
        return false;
    }
}
