/*
    Copyright (c) 2013 Boldizsár Lipka <lipkab@zoho.com>

    This file is part of the Apertium Visual Rule Editor.

    Apertium Visual Rule Editor is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Apertium Visual Rule Editor is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Apertium Visual Rule Editor.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "node.h"
#include "config.h"
#include <QDebug>

Property::Property(const QString &fullName, const QString &value)
    : name_()
    , prefix_()
    , value_(value)
{
    int spos = fullName.lastIndexOf("/");
    if (spos != -1) {
        name_ = fullName.right(fullName.size() - spos - 1);
        prefix_ = fullName.left(spos);
    }
}

Property::Property()
    : name_()
    , prefix_()
    , value_()
{}

void Property::setFullName(const QString &str)
{
    if (str == (prefix_ + "/" + name_)) {
        return;
    }

    int spos = str.lastIndexOf("/");
    if (spos != -1) {
        name_ = str.right(str.size() - spos - 1);
        prefix_ = str.left(spos);
    }

    emit fullNameChanged(str);
}

void Property::setName(const QString &str)
{
    if (str == name_) {
        return;
    }
    name_ = str;

    emit fullNameChanged(prefix_ + "/" + name_);
}

void Property::setValue(const QString &str)
{
    if (str == value_) {
        return;
    }
    value_ = str;
    emit valueChanged(str);
}

class NodeXmlHandler : public QXmlDefaultHandler
{
public:
    NodeXmlHandler()
        : QXmlDefaultHandler()
        , stack_()
        , root_(NULL)
    {}

    virtual bool startElement(const QString &namespaceURI, const QString &localName, const QString &qName, const QXmlAttributes &atts);
    virtual bool endElement(const QString &namespaceURI, const QString &localName, const QString &qName);
    virtual bool characters(const QString &ch);

    Node* root() const { return root_; }

private:
    static QString stripIgnorableWS(const QString &str);

    QList<Node*> stack_;
    Node* root_;
};


bool NodeXmlHandler::startElement(const QString &, const QString &, const QString &qName, const QXmlAttributes &atts)
{
    Node* n = new Node(qName);

    if (root_ != NULL) {
        Node* parent = stack_.back();
        parent->addChild(n);
    } else {
        root_ = n;
    }

    for (int i = 0; i<atts.count(); i++) {
        Property *prop = new Property(n->name() + "/" + atts.qName(i), atts.value(i));
        n->addProperty(prop);
    }

    stack_.append(n);

    return true;
}

bool NodeXmlHandler::endElement(const QString &, const QString &, const QString &)
{
    stack_.removeLast();
    return true;
}

bool NodeXmlHandler::characters(const QString &ch)
{
    stack_.back()->appendCData(stripIgnorableWS(ch));
    return true;
}

QString NodeXmlHandler::stripIgnorableWS(const QString &str)
{
    QString res = str;
    for (int pos = 0; pos<res.size(); pos++) {
        if (res[pos] != '\n') continue;
        while (res[pos].isSpace()) {
            res.remove(pos,1);
        }
    }

    return res;
}

Node* readXmlIntoNode(const QXmlInputSource &xml)
{
    QXmlSimpleReader reader;
    NodeXmlHandler handler;
    reader.setContentHandler(&handler);
    reader.parse(&xml);

    return handler.root();
}

Node::~Node()
{
    foreach(Node* ch, children_) {
        delete ch;
    }
}

void Node::insertChild(Node *ch, int index)
{
    if (!children_.contains(ch)) {
        ch->parent_ = this;
        children_.insert(index, ch);
    }
}

QString Node::toXml() const
{
    QString res;
    res += "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n";
    res += "<" + name_;
    foreach (Property* prop, properties_) {
        res += " " + prop->name() + "=\"" + prop->value() + "\"";
    }
    if (cdata_.isEmpty() && children_.isEmpty()) {
        res += " />\n";
        return res;
    }
    res += ">\n";

    if (!cdata_.isEmpty()) {
        res += "  " + cdata_ + '\n';
    }

    foreach (Node* ch, children_) {
        res += ch->toXmlIndented(1);
    }
    res += "</" + name_ + ">\n";

    return res;
}

QString Node::toXmlIndented(int level) const
{
    QString res;
    QString indent;
    for (int i = level; i>0; i--) {
        indent += "  ";
    }
    res += indent + "<" + name_;
    foreach (Property* prop, properties_) {
        res += " " + prop->name() + "=\"" + prop->value() + "\"";
    }
    if (cdata_.isEmpty() && children_.isEmpty()) {
        res += " />\n";
        return res;
    } else {
        res += ">\n";
    }

    if (!cdata_.isEmpty()) {
        res += indent + "  " + cdata_ + '\n';
    }

    foreach (Node* ch, children_) {
        res += ch->toXmlIndented(level+1);
    }
    res += indent + "</" + name_ + ">\n";

    return res;
}
