/*
    Copyright (c) 2013 Boldizsár Lipka <lipkab@zoho.com>

    This file is part of the Apertium Visual Rule Editor.

    Apertium Visual Rule Editor is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Apertium Visual Rule Editor is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Apertium Visual Rule Editor.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef NODE_H
#define NODE_H

#include <QString>
#include <QList>
#include <QXmlInputSource>
#include <QXmlDefaultHandler>

class Property : public QObject
{
    Q_OBJECT
public:
    Property(const QString& fullName, const QString& value);
    Property();

    virtual ~Property() {}

    QString fullName() const { return prefix_ + "/" + name_; }
    void setFullName(const QString& str);

    QString name() const { return name_; }
    void setName(const QString& str);

    const QString& value() const { return value_; }
    void setValue(const QString& str);

signals:
    void valueChanged(QString);
    void fullNameChanged(QString);

private:
    Property(const Property&);
    QString name_;
    QString prefix_;
    QString value_;
};

class Node : public QObject
{
    Q_OBJECT
public:
    typedef QList<Node*>::Iterator ChildIterator;
    typedef QList<Node*>::ConstIterator ConstChildIterator;
    typedef QList<Property*>::Iterator PropertyIterator;
    typedef QList<Property*>::ConstIterator ConstPropertyIterator;

    Node(const QString& name, Node* parent = NULL)
        : name_(name)
        , properties_()
        , children_()
        , cdata_()
        , parent_(parent)
    {}

    Node(Node* parent = NULL)
        : name_()
        , properties_()
        , children_()
        , cdata_()
        , parent_(parent)
    {}

    virtual ~Node();

    const QString& name() const { return name_; }
    void setName(const QString& str) { name_ = str; }

    const QList<Property*>& properties() const { return properties_; }

    QList<Node*>& children() { return children_; }
    const QList<Node*>& children() const { return children_; }


    const QString& cdata() const { return cdata_; }
    void setCData(const QString& str) { cdata_ = str; }
    void appendCData(const QString& str) { cdata_ += str; }

    Node* parentNode() const { return parent_; }

    QString toXml() const;

public slots:
    void addChild(Node* ch) { if (!children_.contains(ch)) { ch->parent_ = this; children_.append(ch); } }
    void insertChild(Node* ch, int index);
    void removeChild(Node* n) { children_.removeOne(n); }
    void addProperty(Property* pr) { if (!properties_.contains(pr)) { properties_.append(pr); } }
    void removeProperty(Property* pr) { properties_.removeOne(pr); }

private:
    const Node& operator =(const Node& n);
    Node(const Node&);

    QString toXmlIndented(int level) const;

    QString name_;
    QList<Property*> properties_;
    QList<Node*> children_;
    QString cdata_;
    Node* parent_;
};

Node* readXmlIntoNode(const QXmlInputSource& src);

#endif // NODE_H
